/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

/*
 * hostcall.cc
 * alex iliev, nov 2002
 */


#ifdef TESTING_HOSTCALL
#include <fstream>
#endif

#include <unistd.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/sccnet.h>
#include <netinet/in.h> /* for htons */


#include <common/sccutils.h>
#include <common/hostcall.h>
#include <common/exceptions.h>
#include <common/xdr_class.h>
#include <common/consts.h>


#include "hostcall.h"


using namespace std;



static host_status_t
send_host_request (ByteBuffer & reply, const ByteBuffer& req);


#ifdef TESTING_HOSTCALL

#include <stdlib.h>
#include <iostream>

int main (int argc, char* argv[]) {

    // just run a clear record fetch

    if (argc < 3) {
	cerr << "Usage: " << argv[0] << " <bucket #> <record #>" << endl;
	exit (EXIT_FAILURE);
    }
    
    record_id id = { atoi(argv[1]),
		     atoi(argv[2]) };
    Record rec;

    host_status_t status = host_fetch_clearrecord (id, rec);
    if (status == STATUS_OK) {
	cout << "The record-->" << rec << "<--" << endl;
    }
    else {
	cerr << "fetch_clearrecord Host error " << STATUS_OK << endl;
    }

    // and now the same encrypted fetch
    encrypted_record encrec;
    status = host_fetch_record (id, encrec);
    if (status == STATUS_OK) {
	cout.write (reinterpret_cast<char*> (encrec.bytes.data()),
		    encrec.bytes.len());
    }
    else {
	cerr << "fetch_record Host error " << STATUS_OK << endl;
    }
    
}

#endif // TESTING_HOSTCALL


//
// the user service functions
//

host_status_t host_fetch_record (const record_id& id,
				 encrypted_record & o_result)
{
    host_status_t answer;

    // serialize param
    XDR_STRUCT(record_id) xdrcall (id);
    ByteBuffer param = xdrcall.encode();
    ByteBuffer result;
    
    answer = invoke_host (HOST_FETCH_RECORD, param, result);

    if (answer == STATUS_OK) {
	o_result.reconstruct (result);
    }

    return answer;
}


host_status_t host_fetch_clearrecord (const record_id& id,
				      Record & o_result)
{
    host_status_t answer;

    // serialize param
    XDR_STRUCT(record_id) xdrcall (id);
    ByteBuffer param = xdrcall.encode();
    ByteBuffer result;
    
    answer = invoke_host (HOST_FETCH_CLEARRECORD, param, result);

    if (answer == STATUS_OK) {
#ifdef TESTING_HOSTCALL
	ofstream ofs ("record-card.bin");
	ofs.write (reinterpret_cast<char*> (result.data()), result.len());
	ofs.close();
#endif
	o_result.reconstruct (result);
    }

    return answer;
}


host_status_t host_write_record (const named_record& rec)
{
    host_status_t answer;

    ByteBuffer param = rec.serialize();
    ByteBuffer result;

    answer = invoke_host (HOST_WRITE_RECORD, param, result);

    return answer;
}



//
// the function which will be used to invoke services on the host
//
host_status_t invoke_host (host_service_id_t service,
			   const ByteBuffer& params,
			   ByteBuffer& o_result)
{
    host_call_buffer callbuf = { service, params.to_xdr() };

    XDR_STRUCT(host_call_buffer) xdrcall (callbuf);
    ByteBuffer tosend = xdrcall.encode();

    // send the request to the host...
    ByteBuffer answer;
    host_status_t status = send_host_request (answer, tosend);
    if (status != STATUS_OK) {
	perror ("Host communication failure");
	return status;
    }

    XDR_STRUCT(host_return_buffer) xdrret;
    xdrret.decode (answer);

    if (xdrret.x.status == STATUS_OK) {
	ByteBuffer result (xdrret.x.result);
	o_result = result;
    }

    return xdrret.x.status;
}



// FIXME: make all these functions  go in an object and get rid of the globals
static int g_host_sock = -1;
static struct sockaddr_scc g_scc_host_addr;

static host_status_t
send_host_request (ByteBuffer & reply, const ByteBuffer& req) {

    host_status_t status = STATUS_OK;
    int rc;
    
    unsigned short
	cardno   = 0,		// the host
	hostport = PIR_HOSTSERV_PORT;

    const size_t bufsize = 2048;
    byte buf[bufsize];
    

    // set up the socket to the host, only the first time
    if (g_host_sock == -1) {
	if ( scc_setup_send (&g_host_sock, &g_scc_host_addr,
			     cardno, hostport) != 0 )
	{
	    status = STATUS_COMM_FAILURE; goto egress;
	}
    }

    // send request
    if ( (rc = sendto (g_host_sock, req.data(), req.len(),
		       0,
		       reinterpret_cast<struct sockaddr *> (&g_scc_host_addr),
		       sizeof (g_scc_host_addr)))
	 != static_cast<int>(req.len()) )
    {
	status = STATUS_COMM_FAILURE; goto egress;
    }


    struct sockaddr_scc sender_addr;
    socklen_t sender_len;
    sender_len = sizeof (sender_addr);
    if ( (rc = recvfrom (g_host_sock, buf, bufsize,
			 0,
			 reinterpret_cast<struct sockaddr *>(&sender_addr),
			 &sender_len))
	 < 0 )
    {
	status = STATUS_COMM_FAILURE; goto egress;
    }

    if (rc == bufsize) {
	fprintf (stderr, "Recvfrom may have truncated due to small buffer!");
    }

    // TODO: maybe do something to check where the packet came from...

    reply = ByteBuffer (new byte[rc], rc);
    memcpy (reply.data(), buf, reply.len());


 egress:
//    if (sock >= 0) close (sock);
    return status;
}
