/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

/*
 * pirserver.cc
 * alex iliev, jan 2003
 * the card-side code to handle PIR requests from outside
 */

#include <string>
#include <vector>
#include <new>

#include <sstream>

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <errno.h>

#include <unistd.h>

#include <sys/socket.h>
#include <sys/sccnet.h>
#include <netinet/in.h> /* for htons */


#include <common/utils.h>
#include <common/sccutils.h>
#include <common/sym_crypto.h>
#include <common/consts.h>
#include <common/exceptions.h>


#include "lib.h"
#include "pirsession.h"
#include "consts.h"


using namespace std;


static const size_t REQSIZE = 1024;




// HACK: ugh, global
char * g_program;

// these two stolen from sccping.c from IBM
static void
error(char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "%s: Error: ", g_program);
    vfprintf(stderr, fmt, ap);
    fputc('\n', stderr);
    va_end(ap);
    exit(EXIT_FAILURE);
}

static void
warning(char *fmt, ...)
{
    va_list ap;

    va_start(ap, fmt);
    fprintf(stderr, "%s: Warning: ", g_program);
    vfprintf(stderr, fmt, ap);
    fputc('\n', stderr);
    va_end(ap);
}


static void reqloop (unsigned short sock, PIRSession & session);


static void
usage(const char* progname)
{
    fprintf(stderr,
	"Usage: %s [-p port] [-d <root directory>]\n", progname);
    exit(EXIT_FAILURE);
}



void out_of_memory () {
    cerr << "Out of memory!" << endl;
}


int main (int argc, char *argv[]) {

    int port = PIR_MAIN_PORT;
    int s;		// the socket descriptor

    string dir = ROOTDIR;


    // from sccping.c
    int opt;
    while ((opt = getopt(argc, argv, "p:d:")) != EOF) {
	switch (opt) {

        case 'p':		// port
	    port = strtol(optarg, NULL, 0);
	    if (port == 0)
		error("invalid port");
	    break;

	case 'd':		//directory for keys etc
	    dir = optarg;
	    break;

	default:
	    usage(argv[0]); 
	}
    }

    g_program = argv[0];


    set_new_handler (out_of_memory);
    
    // setup the socket
    if ( (s = scc_setup_server_sock (port)) < 0 ) {
	error("socket: %s", strerror(errno));
    }

    
    // initialize a PIR session object
    // for this, need to get a translation vector, and two keys
    // for now, fetch them off the host file system!
    byte key[DES3_KEY_SIZE], mackey[DES3_KEY_SIZE];
    ByteBuffer
	keybuf    (key,    DES3_KEY_SIZE, ByteBuffer::no_free),
	mackeybuf (mackey, DES3_KEY_SIZE, ByteBuffer::no_free);
    loadkeys (key,    dir + DIRSEP + KEYFILE,
	      mackey, dir + DIRSEP + MACKEYFILE);
    
    clog << "About to load the shuffle" << endl;
    
    vector<index_t> V;
    load_shuffle (V, dir + DIRSEP + SHUFFLEFILE);

    PIRSession session (keybuf, mackeybuf, &V, BUCKETSIZE);

    clog << "Starting the request loop" << endl;
    
    reqloop (s, session);
    

}


static void reqloop (unsigned short sock, PIRSession & session) {

    char reqbuf[REQSIZE];

    struct sockaddr_scc scc;
    socklen_t scclen;
    int rc;
    
    while(true) {

	/* get request */
    	scclen = sizeof(scc);
	if ((rc = recvfrom
	     (sock, reqbuf, REQSIZE-1, 0,
	      reinterpret_cast<struct sockaddr *>(&scc), &scclen)) < 0)
	{
	    close (sock);
	    error("recvfrom: %s", strerror(errno));
	}

	if (rc == static_cast<int>(REQSIZE)) {
	    // truncation??
	    // TODO
	}

	reqbuf[rc] = '\0';

	string name = reqbuf;
	Record rec;
	clog << "Retrieving name " << name << endl;
	try {
	    rec = session.find_record (name);
	}
	catch (const exception& ex) {
	    cerr << ex.what() << endl;
	    continue;
	}

	ostringstream os;
	os << rec;

	if ( (rc = sendto
	      (sock, os.str().c_str(), os.str().length(), 0,
	       reinterpret_cast<struct sockaddr *>(&scc), scclen))
	     != static_cast<int>(os.str().length()) )
	{
	    warning ("sending reply: %s", strerror(errno));
	}

    }
}
