/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

// shuffle.cc
// alex iliev jan 2003
// routines for PIR database shuffling, in the Shuffler class


#include <vector>
#include <fstream>
#include <list>
#include <functional>
#include <exception>

#include <unistd.h>		// for getopt()
#include <time.h>		// for time()


#include <common/exceptions.h>
#include <common/record.h>
#include <common/sym_crypto.h>
#include <common/consts.h>

#include "lib.h"
#include "hostcall.h"
#include "consts.h"
#include "shuffle.h"



using namespace std;



void out_of_memory () {
    cerr << "Out of memory!" << endl;
}


static void
usage(const char* progname)
{
    fprintf(stderr,
	    "Usage: %s [-p <card server port>] [-d <root dir for keys etc.>]\n",
	    progname);
    exit(EXIT_FAILURE);
}



int main (int argc, char * argv[]) {


    unsigned short host_serv_port = PIR_HOSTSERV_PORT;
    string dir = ROOTDIR;
    
    byte key[DES3_KEY_SIZE], mackey[DES3_KEY_SIZE];

    ByteBuffer keybuf (key, DES3_KEY_SIZE, ByteBuffer::no_free),
	mackeybuf (mackey, DES3_KEY_SIZE, ByteBuffer::no_free);

    set_new_handler (out_of_memory);

    int opt;
    while ((opt = getopt(argc, argv, "p:d:")) != EOF) {
	switch (opt) {

        case 'p':		// port
	    host_serv_port = strtol(optarg, NULL, 0);
	    if (host_serv_port == 0) {
		cerr << argv[0] << ": Invalid port" << endl;
		exit (EXIT_FAILURE);
	    }
	    break;

	case 'd':		//directory for keys etc
	    dir = optarg;
	    break;

	default:
	    usage(argv[0]); 
	}
    }

    

    loadkeys (key,    dir + DIRSEP + KEYFILE,
	      mackey, dir + DIRSEP + MACKEYFILE);


    // DBSIZE and BUCKETSIZE are for now in lib.h
    Shuffler shuffler (keybuf, mackeybuf, DBSIZE, BUCKETSIZE);
    
    try {
	shuffler.generate_shuffle();
	shuffler.write_shuffle (dir + DIRSEP + SHUFFLE_FILE);
	shuffler.shuffle ();
    }
    catch (const exception& ex) {
	cerr << "Exception from shuffle!" << endl
	     << ex.what() << endl;
	return EXIT_FAILURE;
    }

}


Shuffler::Shuffler (const ByteBuffer & key, const ByteBuffer & mackey,
		    size_t N, size_t bucksize)
    : _key(key),
      _mackey(mackey),
      _enc(key),
      _mac(mackey),
      N(N),
      _bucket_size (bucksize)
{}



void Shuffler::generate_shuffle () {
    // prepare the vector, so it contains [0..N-1]
    V.reserve (N); V.resize (N);
    for (index_t i=0; i < N; i++) V[i] = i;

    // FIXME: reconsider the randomness source here
    srandom (time(NULL));
    
    for (int i = N-1; i > 0; i--) {
	int j = random () % (i+1);
	swap (V[i], V[j]);
    }
}



template<class T> struct print : public unary_function<T, void>
{
    void operator() (const T& x)
	{ os << x << endl; }

    print(ostream& out) : os(out) {}
    ostream& os;
};


void Shuffler::write_shuffle (const string& filename)
    throw (std::ios::failure)
{
    ofstream fos (filename.c_str());
    
    // actaully we need to save a different form of this vector, such that
    //        V'[i] is the shuffled position of record i
    // right now, i is the shuffled position of record V[i]
    vector<index_t> V_;
    transpose_vector (V_, V);

    for_each (V_.begin(), V_.end(), print<index_t> (fos));
}


void Shuffler::shuffle ()
    throw (host_exception, crypto_exception)
{
    // tactic:
    // perform the asonov shuffling procedure, treating *buckets* as units
    //     to shuffle

    
    // now, record V[i] should go into position i

    // counting variables:
    // g: the current position we're trying to fill in the shuffled DB
    // h: the record being considered for this position
    for (unsigned g = 0; g < N; g++) {

	Bucket saved;
	
	for (unsigned h = 0; h < N; h++) {

	    Bucket temp;
//	    clog << "Fetching bucket " << h << endl;
	    fetch_bucket (h, temp, _bucket_size);

	    if (h == V[g]) {
		// g is the new position for record h
		saved = temp;	// FIXME: a timing attack possible through
				// selective execution of this line?

		clog << "will write bucket " << h << " to pos " << g << endl;
	    }

	}

	write_bucket (g, saved);

    }
		
}

	
	

//
// retrieve a whole bucket, in the clear.
//
void Shuffler::fetch_bucket (index_t n, Bucket & o_bucket, size_t bucksize)
    throw (host_exception)
{

    Record current;
    host_status_t status = STATUS_OK;

    // just fetch all the records in this buckets, ie until an error comes

    // FIXME: should somehow better be able to tell how many records in a
    // bucket, esp. since it should be the same across all buckets.

    for (index_t i = 0; i < _bucket_size; i++) {
//	clog << "Fetching bucket " << n << " record " << i << endl;
	
	record_id id = {n, i};
	status = host_fetch_clearrecord (id, current);
	if (status != STATUS_OK) {
	    throw host_exception ("Error on record fetch ", status);
	}

// 	clog << "got a record! " << current.attributes->attr.name
// 	     << " = " << current.attributes->attr.value << endl;
	
	o_bucket.records.push_back (current);
    }
}



//
// write a bucket of records to the host, encrypting and maccing in the process
//

void Shuffler::write_bucket (index_t n, const Bucket & bucket)
    throw (crypto_exception, host_exception)
{

    // tactic: need to marshall, encrypt and mac each record, then write the
    // ciphertext out

    host_status_t status;
    list<Record>::const_iterator it;
    int i = 0;

    for (it = bucket.records.begin();
	 it != bucket.records.end();
	 it++, i++)
    {
	ByteBuffer cleartext = it->serialize();

	ByteBuffer ciphertext = _enc.encrypt (cleartext);
	ByteBuffer mac = _mac.genmac (ciphertext);

	ByteBuffer tosend = addmac (ciphertext, mac);


	record_id id = {n, i};
	status = host_write_record (named_record (id, tosend));
	if (status != STATUS_OK) {
	    throw host_exception ("Writing a record during shuffle", status);
	}
    }
    
}




// transpose is used a bit loosely here
// we just want to set V_[V[i]] = i for all indices i
void transpose_vector (vector<index_t> & V_, const vector<index_t>& V) {

    V_.resize (V.size());

    for (unsigned i = 0; i < V.size(); i++) {
	V_[V[i]] = i;
    }
}



