/*
 * ** PIR Private Directory Service prototype
 * ** Copyright (C) 2002 Alexander Iliev <iliev@nimbus.dartmouth.edu>
 * **
 * ** This program is free software; you can redistribute it and/or modify
 * ** it under the terms of the GNU General Public License as published by
 * ** the Free Software Foundation; either version 2 of the License, or
 * ** (at your option) any later version.
 * **
 * ** This program is distributed in the hope that it will be useful,
 * ** but WITHOUT ANY WARRANTY; without even the implied warranty of
 * ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * ** GNU General Public License for more details.
 * **
 * ** You should have received a copy of the GNU General Public License
 * ** along with this program; if not, write to the Free Software
 * ** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * */

/*
 * host/hostcall.cc
 * alex iliev, nov 2002
 */

#include <fstream>
#include <iostream>

#include <sstream>

// these 3 for stat() and mkdir()
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#include <time.h>		// for nanosleep

#include <common/utils.h>
#include <common/hostcall.h>
#include <common/xdr_class.h>
#include <common/record.h>
#include <common/comm_types.h>

#include "hostcall.h"


using namespace std;


static const char CRYPTROOT[] = "/home/sasho/work/shib/store/crypt",
    CLEARROOT[] = "/home/sasho/work/shib/store/clear";


static const string DIRSEP = "/";



    
//
// little helpers
//
static ByteBuffer
fetch_record (index_t bucket, index_t record, const string& rootdir)
    throw (ios::failure);


static void
fetch_clear_record (index_t bucket, index_t record, Record & o_rec)
    throw (std::ios_base::failure);


static bool
dir_exists (const string& name) throw (std::ios_base::failure);




//
// the handling functions, same prototypes as in the card (just accidental)
// keep them out of sight of other code!
//
static host_status_t host_fetch_record (const record_id& id,
					encrypted_record & o_result);

static host_status_t host_fetch_clearrecord (const record_id&,
					     Record & o_result);

static host_status_t host_write_record (const named_record&);



#ifdef TESTING_HOSTCALL

int main () {
    
    Record rec;

    try {

	for (index_t i = 0; i < 3000; i++) {
	    fetch_clear_record (i, 0, rec);
	    cout << "second attrib: " << rec.attributes.front().name
		 << " = " << rec.attributes.front().value << endl;
	}

    }
    catch (const ios::failure & ex) {
	cerr << "Exception: " << ex.what() << endl;
	exit (1);
    }


}

#endif // TESTING_HOSTCALL


ByteBuffer host_handle_request (const ByteBuffer& req) {

    XDR_STRUCT(host_call_buffer) xdrreq;
    xdrreq.decode (req);

    host_service_id_t servid = xdrreq.x.service_id;

    ByteBuffer answer;
    host_status_t status;
    
    switch (servid) {
    case HOST_FETCH_RECORD:
    {
	XDR_STRUCT(record_id) xdrparam;
	encrypted_record encrec;

	// xdrreq.x.params is of type ByteBuffer_x
	ByteBuffer paramsbuf (xdrreq.x.params);
	xdrparam.decode (paramsbuf);
	
	// do the work
	status = host_fetch_record (xdrparam.x, encrec);

	if (status == STATUS_OK) {
	    answer = encrec.serialize();
	}
    }
    break;
	
    case HOST_FETCH_CLEARRECORD:
    {
	XDR_STRUCT(record_id) xdrparam;
	Record rec;

	xdrparam.decode (xdrreq.x.params);
	
	// do the work
	status = host_fetch_clearrecord (xdrparam.x, rec);

	if (status == STATUS_OK) {
	    answer = rec.serialize();

	    ofstream ofs ("record-host.bin");
	    ofs.write (reinterpret_cast<char*> (answer.data()), answer.len());
	    ofs.close();
	}
    }
    break;

    case HOST_WRITE_RECORD:
    {
	// do the work
	named_record r (ByteBuffer (xdrreq.x.params)); // deserialize
	status = host_write_record (r);

	// 'answer' will be left empty here
    }
    break;

    }

    // construct the return buffer

    // damn - the buffer we return will actually be deallocated at the end of
    // this function!
    host_return_buffer retbuf = { status, answer.to_xdr() };
    
    XDR_STRUCT(host_return_buffer) xdrret (retbuf);
    return xdrret.encode();
}






inline static void delay () {
    return;
#if 0
    struct timespec t;
    t.tv_sec = 0;
    t.tv_nsec = 100000000;	// 100,000,000 nsec = 1/10 second
    nanosleep (&t, NULL);
#endif
}


static host_status_t host_fetch_record (const record_id& id,
					encrypted_record& o_result)
{
    //
    // for now we can organize the records in directories so we will fetch
    // file $ROOT/bucket/record
    //

    delay ();			// to simulate the slow host-card transfer!
    
    try {
	o_result.bytes = fetch_record (id.bucket, id.record, CRYPTROOT);
    }
    catch (const ios::failure & ex) {
	return STATUS_IO_ERROR;
    }
    
    return STATUS_OK;
}



static host_status_t host_fetch_clearrecord (const record_id& id,
					     Record & o_result)
{
    
    Record rec;
    try {
	fetch_clear_record (id.bucket, id.record, o_result);
    }
    catch (const ios::failure & ex) {
	clog << "fetch_clear_record failed: " << ex.what() << endl;
	return STATUS_IO_ERROR;
    }

    return STATUS_OK;
}



static host_status_t host_write_record (const named_record& rec)
{
    // may have to create some directories!
    // start from the root, assume it exists

    delay ();
    
    ostringstream filename;
    filename << CRYPTROOT << DIRSEP << rec.id.bucket;

    // FIXME: dir_exists may throw an exception!
    if (!dir_exists (filename.str())) {
	if ( mkdir (filename.str().c_str(), 0777) != 0 ) {
	    return STATUS_DIR_ERROR;
	}
    }

    filename << DIRSEP << rec.id.record;

    ofstream fos (filename.str().c_str());

    fos.write (reinterpret_cast<char*> (rec.rec_bytes.data()),
	       rec.rec_bytes.len());
    if (fos.fail()) {
	return STATUS_IO_ERROR;
    }

    return STATUS_OK;

}




//
// little helpers
//

bool dir_exists (const string& name) throw (std::ios_base::failure) {

    struct stat st;
    memset (&st, 0, sizeof(st));

    clog << "Checking dir " << name << endl;
    
    if (stat (name.c_str(), &st) != 0) {
	return false;
    }

    if (S_ISDIR(st.st_mode)) {
	return true;
    }

    // and now?
    // FIXME: invent an excpetion hier. especially for these host calls
    throw std::ios_base::failure ("Dir name exists but is not a directory");
}



ByteBuffer
fetch_record (index_t bucket, index_t record, const string& rootdir)
    throw (ios::failure)
{

    ostringstream filename_str;
    filename_str << rootdir << DIRSEP << bucket << DIRSEP << record;
    string filename = filename_str.str();

    string contents;
    FILE * fh = fopen (filename.c_str(), "r");
    if (!fh) {
	throw ios::failure ("File open failed: " +
			    string (strerror(errno)));
    }

    readfile (fh, contents);
    fclose (fh);

    ByteBuffer bytes (new byte[contents.length()], contents.length());
    contents.copy (reinterpret_cast<char*> (bytes.data()), bytes.len());

    return bytes;
}



static void
open_rec_file (index_t bucket, index_t record, const string& root,
	       ifstream & o_ifs)
    throw (ios::failure)
{
    ostringstream filename;
    
    filename << CLEARROOT << DIRSEP << bucket << DIRSEP << record;

//     if (!dir_exists (filename.str())) {
// 	throw std::ios_base::failure ("Not a directory: " + filename.str());
//     }

//     filename << DIRSEP << record;

    o_ifs.open (filename.str().c_str());
    if (!o_ifs) {
	cerr << "Failed to open " << filename.str() << endl;
    }
}


void
fetch_clear_record (index_t bucket, index_t record,
		    Record & o_rec)
    throw (ios::failure)
{
    ifstream ifs;
    
    // FIXME: this exception stuff doesnt seem to work as advertised. abandon
    // for now
    ifs.exceptions (ifstream::badbit);

    open_rec_file (bucket, record, CLEARROOT, ifs);
    if (!ifs) {
	throw ios::failure ("Failed to open record file for reading");
    }

    
    // the name
    getline (ifs, o_rec.name);
    
    // and now the attributes
    RecordAttr current;
    
    while (getline (ifs, current.name) && getline (ifs, current.value)) {
	o_rec.attributes.push_back (current);
    }
}
